/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
 **/
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic.smx.components.tcp;

import java.net.InetSocketAddress;

import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.mina.common.ByteBuffer;
import org.apache.mina.common.ConnectFuture;
import org.apache.mina.common.DefaultIoFilterChainBuilder;
import org.apache.mina.common.IoConnector;
import org.apache.mina.common.IoSession;
import org.apache.mina.common.RuntimeIOException;
import org.apache.mina.common.WriteFuture;
import org.apache.mina.filter.LoggingFilter;
import org.apache.mina.filter.SSLFilter;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.nio.SocketConnector;
import org.apache.servicemix.common.DefaultComponent;
import org.apache.servicemix.common.ServiceUnit;
import org.apache.servicemix.common.endpoints.ProviderEndpoint;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.apache.servicemix.jbi.util.MessageUtil;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.spagic.smx.components.tcp.codec.TCPMsgCodecFactory;
import org.spagic.smx.components.tcp.ssl.TCPBCSSLContextFactory;
import org.spagic.smx.components.tcp.util.NMUtils;

/**
 * Implement the "TCPClient"
 * 
 * @org.apache.xbean.XBean element="provider"
 */
public class TCPProviderEndpoint extends ProviderEndpoint implements
		TCPEndpointType {

	private static final Log log = LogFactory.getLog(TCPProviderEndpoint.class);
	// private static final String ERROR_MSG_KEY = "ERROR_MSG_KEY".intern();
	private static final double retryPower = 2.0;
	private TCPBCConfig config;
	private TCPMarshaller marshaller;
	private ConnectFuture future;

	public TCPProviderEndpoint() {
	}

	private TCPMarshaller getMarshaller() {
		if (marshaller == null) {
			marshaller = new TCPMarshaller(config);
		}
		return marshaller;
	}

	public TCPProviderEndpoint(ServiceUnit serviceUnit, QName service,
			String endpoint) {
		super(serviceUnit, service, endpoint);
	}

	public TCPProviderEndpoint(DefaultComponent component,
			ServiceEndpoint endpoint) {
		super(component, endpoint);
	}

	public void validate() throws DeploymentException {
		super.validate();
	}

	protected void processInOnly(MessageExchange exchange, NormalizedMessage in)
			throws Exception {
		// TODO now you do not respond in case of IN-OUT
		processInOut(exchange, in, null);
	}

	/**
	 * Use MINA to open a comunication to a TCPServer end.
	 */
	protected void processInOut(MessageExchange exchange, NormalizedMessage in,
			NormalizedMessage out) throws Exception {

		log.debug("processInOut->start");

		log.debug("processInOut->Creating IOConnector");
		IoConnector connector = new SocketConnector();

		log.debug("processInOut->IOConnector Creaing Filter Chain");
		DefaultIoFilterChainBuilder chain = connector.getDefaultConfig()
				.getFilterChain();

		if (config.isUseSSL()) {
			log
					.debug("processInOut->SSL Supporr Required configure SSL In Chain");
			addSSLSupport(chain);
		}
		log.debug("processInOut-> Add The Codec Protocol Filter ");
		chain.addLast("codec", new ProtocolCodecFilter(new TCPMsgCodecFactory(
				config)));

		if (log.isDebugEnabled()) {
			addLogger(chain);
		}

		if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode())
				|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config
						.getConnectionMode())) {

			log
					.debug("processInOut-> CONN_CODE_MANTAIN -> SYNCHRONIZATION REQUIRED on Object ");
			synchronized (this) {
				log
						.debug("processInOut-> call sendTCPMessage : Entered in Synchnronized Code Block ");
				sendTCPMessage(connector, exchange, in, out);
				log
						.debug("processInOut-> after send : Finish Synchnronized Code Block ");
			}

		} else {
			log
					.debug("processInOut-> Synchronization NOT REQUIRED : call send ");
			sendTCPMessage(connector, exchange, in, out);
			log.debug("processInOut-> after sendTCPMessage ");
		}

		// log.debug("processInOut--> Verify Configuration ");
		// log.debug("**** TCPBCConfig.CONN_MODE_MAINTAIN -> " +
		// TCPBCConfig.CONN_MODE_MAINTAIN);
		// log.debug("**** config.getConnectionMode() -> " +
		// config.getConnectionMode());
		// log.debug("****
		// (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode()) ->
		// " +
		// TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode()));
		// synchronized (this) {
		// log.debug("processInOut-> call sendTCPMessage : Entered in
		// Synchnronized Code Block ");
		// sendTCPMessage(connector, exchange, in, out);
		// log.debug("processInOut-> after send : Finish Synchnronized Code
		// Block ");
		// }
		// log.debug("processInOut-> stop");

	}

	private IoSession getTCPSession(ConnectFuture future) throws Exception {
		// can throw exception
		IoSession aSession = null;
		long previousDelayTime = config.getRetryDelay();

		log.debug("sendTCPMessage::getTCPSession Getting Session");

		log.debug("sendTCPMessage::getTCPSession Future Is Ready -> ["
				+ future.isReady() + "]");
		if (!future.isConnected()) {
			// log.debug("sendTCPMessage::getTCPSession Future Is Not Connected
			// -> Must JOIN");
			// future.join();
			// log.debug("JOIN OK");
		}

		// -> Ciclo For
		for (int retry = 0; retry <= config.getRetryNumber(); retry++) {
			try {
				log
						.debug("sendTCPMessage::getTCPSession :: Getting TCP Session Retry ["
								+ retry + "] Getting session");
				aSession = future.getSession();
			} catch (RuntimeIOException rIOe) {
				log.warn("sendTCPMessage::Runtime Exception", rIOe);
			}
			if (aSession != null) {
				log
						.debug("sendTCPMessage::getTCPSession :: Getting TCP Session Retry ["
								+ retry + "] Getted OK ");
				log.debug("sendTCPMessage::getTCPSessionSession Is Closing -> "
						+ aSession.isClosing());
				log
						.debug("sendTCPMessage::getTCPSessionSession Is Connected -> "
								+ aSession.isConnected());
				if (!aSession.isClosing() && aSession.isConnected()) {
					log.debug("sendTCPMessage::getTCPSession:: Session OK");
					return aSession;
				}
			} else {
				log
						.debug("sendTCPMessage::getTCPSession Session is NOT VALID ");
				if (retry <= config.getRetryNumber()) {
					long delay = getRetryWaitTime(previousDelayTime);
					log
							.debug("sendTCPMessage::getTCPSession:: Sleeping For Retry Wait Time"
									+ delay);
					if (delay == -1) {
						return null;
					}
					log
							.debug("sendTCPMessage::getTCPSession::Retring in: (ms) "
									+ delay);
					previousDelayTime = delay;
					try {
						Thread.sleep(delay);
					} catch (InterruptedException ie) {
						log.error("Error waiting for retry", ie);
					}
				}
			}

		}
		return aSession;

	}

	private void sendTCPMessage(IoConnector connector,
			MessageExchange exchange, NormalizedMessage in,
			NormalizedMessage out) throws Exception {

		log.debug("sendTCPMessage(connector,exchange,in,out)");
		// connect to the TCPServer using Retry policies
		ConnectFuture future = getConnection(connector, in);
		if (future == null) {
			// no connection is possible
			log.warn("sendTCPMessage::connect future is null");
			return;
		}

		IoSession session = getTCPSession(future);

		if (session == null) {
			String remoteHost = getDynamicRemoteHost(in);
			Integer remotePort = getDynamicRemotePort(in);

			log.warn("Error creating connection host=" + remoteHost + ", port="
					+ remotePort);
			Fault fault = exchange.createFault();
			Element faultElem = DocumentHelper.createElement("Fault");
			faultElem.setText("Error creating connection host=" + remoteHost
					+ ", port=" + remotePort);
			fault.setContent(new StringSource(faultElem.asXML()));
			MessageUtil.transfer(fault, out);
			return;
		}

		if (session.isConnected()) {
			log.debug("Session Is Connected");
			try {
				boolean received = false;
				session.setAttribute(
						TCPClientProtocolHandler.WAIT_RESPONSE_KEY,
						Boolean.FALSE);
				for (int numRetry = 0; numRetry < config.getRetryCount()
						&& !received; numRetry++) {

					// write the message
					Object msgObj = NMUtils.retrieveMessageInEnvelope(in,
							config.getInNmEnvelope() + "/text()");
					if (config.isBase64decode()) {
						byte[] decodedB = Base64.decodeBase64((byte[]) msgObj);
						msgObj = decodedB;
					}

					log.debug("sendTCPMessage::Retry=" + numRetry
							+ " -> Message to send  " + msgObj);

					/*
					 * if (log.isDebugEnabled()) { log.debug("Message to send
					 * (retry="+numRetry+"): "+msgObj); log.debug("Checks before
					 * send"); Boolean waitResponseKey =
					 * (Boolean)session.getAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY);
					 * log.debug("Is there response message? "+waitResponseKey);
					 * if (waitResponseKey) { log.debug("Response message: "+
					 * session.getAttribute(TCPClientProtocolHandler.MESSAGE_INSESSION_KEY)); } }
					 */
					log.debug("sendTCPMessage::Retry=" + numRetry
							+ " -> Before WriteObject");
					WriteFuture wFuture = session.write(msgObj);
					wFuture.join();
					log.debug("sendTCPMessage::Retry=" + numRetry
							+ " -> After wFuture.join()");

					if (exchange instanceof InOnly
							|| exchange instanceof RobustInOnly) {
						// terminate the exchange
						done(exchange);
					}

					try {
						log.debug("sendTCPMessage::Retry=" + numRetry
								+ " -> Waiting For Response ");
						received = waitForResponse(session);
						log.debug("sendTCPMessage::Retry=" + numRetry
								+ " -> Response Arrived ");
						Object response = session
								.getAttribute(TCPClientProtocolHandler.MESSAGE_INSESSION_KEY);
						if (received && (response instanceof ByteBuffer)) {
							log
									.debug("sendTCPMessage::Retry="
											+ numRetry
											+ " -> Transforming Response To Normalized Message ");
							getMarshaller().toNMS(exchange, out,
									(ByteBuffer) response);
							log.debug("sendTCPMessage::Retry=" + numRetry
									+ " -> OK ");
							session
									.setAttribute(TCPClientProtocolHandler.MESSAGE_INSESSION_KEY);
						} else {
							log
									.warn("sendTCPMessage::Retry="
											+ numRetry
											+ " -> Unable to handle tcp response of type: "
											+ ((response != null) ? response
													.getClass().getName()
													: response));
						}
					} catch (InterruptedException ie) {
						log.error("sendTCPMessage::Retry=" + numRetry
								+ " -> Error waiting for a response", ie);
					} catch (Throwable t) {
						log.error("sendTCPMessage::Retry=" + numRetry
								+ " -> Uncatched Exception Occurred", t);
					} finally {
						// prepare session for next request response exchange
						session
								.setAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY);
					}
				}
				if (!received) {
					log
							.warn("sendTCPMessage::No response received after retry policy");
					Fault fault = exchange.createFault();
					Element faultElem = DocumentHelper.createElement("Fault");
					faultElem.setText("No response received from TCP server");
					log.error("sendTCPMessage::Fault Content -> "
							+ new StringSource(faultElem.asXML()).toString());
					fault.setContent(new StringSource(faultElem.asXML()));
					MessageUtil.transfer(fault, out);
				}
			} finally {
				closeConnection(future);
			}
		} else {
			log.warn("sendTCPMessage::Session not created!");
		}

	}

	private void closeConnection(ConnectFuture future) {

		if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode())
				|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config.getConnectionMode())) {
			log
					.debug("closeConnection::Endpoint Configured To Mantain Session ");
			log.debug("Nothing to do on future and session");
		} else {
			log.debug("closeConnection::Endpoint Configured To Not Mantain Session ");
			log.debug("closeConnection::Endpoint Configured To Not Mantain Session close session and future");
			future.getSession().close();
			future = null;
		}
	}

	private boolean waitForResponse(IoSession session)
			throws InterruptedException {
		Long cycleNumberL = config.getResponseTimeout() / 500;
		int cycleNumber = cycleNumberL.intValue();
		log.info("waitForResponse::Start Polling Loop......");
		for (int j = 0; j < cycleNumber; j++) {
			Thread.sleep(500);
			log.info("waitForResponse::Loop Poll IoSession " + (j + 1) * 500);
			Boolean received = (Boolean) session
					.getAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY);
			if (received.booleanValue()) {
				log.info("waitForResponse::Received after " + (j + 1) * 500);
				return true;
			}
		}
		log.info("waitForResponse::End Polling Loop......");
		Boolean received = (Boolean) session
				.getAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY);
		if (received.booleanValue()) {
			log
					.info("waitForResponse::After End of Polling Loop Returning True......");
			return true;
		}
		log
				.info("waitForResponse::After End of Polling Loop Returning False......");
		return false;
		/*
		 * for( int j = 0; j < 1; j ++ ) { Thread.sleep( 500 ); Boolean received =
		 * (Boolean)session
		 * .getAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY); if(
		 * received.booleanValue() ) { return true; } Thread.sleep(
		 * config.getResponseTimeout() ); //check if in the meanwhile the
		 * response arrived received = (Boolean)session
		 * .getAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY); return
		 * received; } return false;
		 */
	}

	private String getDynamicRemoteHost(NormalizedMessage in) {
		String remoteHost = (String) in
				.getProperty(TCPBCConfig.TCP_DESTINATION_HOST);
		if (remoteHost == null) {
			remoteHost = config.getRemoteHost();
		} else {
			log.debug("Remote Host URI overridden: " + remoteHost);
		}
		return remoteHost;
	}

	private Integer getDynamicRemotePort(NormalizedMessage in) {
		Integer remotePort = (Integer) in
				.getProperty(TCPBCConfig.TCP_DESTINATION_PORT);
		if (remotePort == null) {
			remotePort = config.getRemotePort();
		} else {
			log.debug("Remote port URI overridden: " + remotePort);
		}
		return remotePort;
	}

	/**
	 * Connect to the TCPServer using retry policies: <br/><br/> From Rapsody
	 * manual:<br/> The Retry Count setting determines how many times the
	 * message should be resent while waiting for a response. This can be set to
	 * 0, to indicate that the message should never be resent, or -1, to
	 * indicate that there is no maximum number of resends. Once the maximum
	 * number of resends has been reached and the timeout occurs, the Fail
	 * Action is executed (which is one of Send Message to Error Queue, Close
	 * Connection, or Send to Error Queue and Close Connection.
	 * 
	 * @param connector
	 * @return
	 */
	private ConnectFuture getConnection(IoConnector connector,
			NormalizedMessage in) {
		long previousDelayTime = config.getRetryDelay();

		if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode())
				|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config
						.getConnectionMode())) {
			if ((future != null) && future.isConnected()) {

				log
						.debug("getConnection::Returning already initiated connection");
				return future;
			}
		}

		for (int retry = 0; retry <= config.getRetryNumber(); retry++) {
			log.debug("getConnection::Retry=" + retry + " start");

			try {
				String remoteHost = getDynamicRemoteHost(in);
				log.debug("getConnection::Retry=" + retry + " Remote Host"
						+ remoteHost);
				Integer remotePort = getDynamicRemotePort(in);
				log.debug("getConnection::Retry=" + retry + " Remote Port"
						+ remotePort);

				log.debug("getConnection::Retry=" + retry + " Before Connect");
				ConnectFuture localFuture = connector.connect(
						new InetSocketAddress(remoteHost, remotePort),
						new TCPClientProtocolHandler(config));

				// connection obtained
				localFuture.join();
				log.debug("getConnection::Retry=" + retry + " After Connect");

				if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config
						.getConnectionMode())
						|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config
								.getConnectionMode())) {
					// I forget the previous instance????
					future = localFuture;
				}
				log.debug("getConnection::Retry=" + retry
						+ " Return local Future");
				return localFuture;
			} catch (Exception e) {
				log.error("Error connecting to TCPServer", e);
				if (retry <= config.getRetryNumber()) {
					long delay = getRetryWaitTime(previousDelayTime);
					if (delay == -1) {
						return null;
					}
					log.debug("...Retring in: (ms) " + delay);
					previousDelayTime = delay;
					try {
						Thread.sleep(delay);
					} catch (InterruptedException ie) {
						log.error("Error waiting for retry", ie);
					}
				}
			}
		}
		return null;
	}

	private long getRetryWaitTime(long previousTime) {
		String retryType = config.getRetryType();
		if (retryType.equals(CommonConfig.RETRY_TYPE_IMMEDIATE)) {
			return 0;
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_LINEAR)) {
			return config.getRetryDelay();
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_EXPONENTIAL)) {
			return (long) Math.pow(previousTime, retryPower);
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_NO_RETRY)) {
			return -1;
		}

		return -1;
	}

	private void addSSLSupport(DefaultIoFilterChainBuilder chain)
			throws Exception {
		// TODO check if you can use always the same SSLFilter
		SSLFilter sslFilter = new SSLFilter(TCPBCSSLContextFactory
				.getInstance(config));
		sslFilter.setUseClientMode(true);
		chain.addLast("SSL", sslFilter);
		log.info("SSL ON");
	}

	public TCPBCConfig getConfig() {
		return config;
	}

	public void setConfig(TCPBCConfig config) {
		this.config = config;
	}

	private void addLogger(DefaultIoFilterChainBuilder chain) throws Exception {
		// we have to log CONNECTION and DATA.
		// TODO CHECK If the logger is added after the codec it receives data
		// already
		// converted to the codec protocol?
		chain.addLast("logger", new LoggingFilter());
		log.info("Logging ON");
	}

	@Override
	public void stop() throws Exception {
		super.stop();
		if (future != null) {
			((TCPClientProtocolHandler) future.getSession().getHandler())
					.dispose();
			future.getSession().close();
		}
	}
}
